"""The Dataiku dataframesource class.
"""
import dataiku
import dataiku.spark as dkuspark
import pyspark
from birgitta.dataframesource import DataframeSourceBase
from pyspark.sql import SQLContext


__all__ = ['DataikuSource']


class DataikuSource(DataframeSourceBase):
    def load(self, spark_session, dataset_name, prefix, **kwargs):
        """When prefix is None use same project as recipe is in."""
        project_key = prefix  # Assume that prefix is a valid project_key
        dataset = dataiku.Dataset(dataset_name, project_key)
        sql_ctx = self.dataiku_sql_ctx()
        return dkuspark.get_dataframe(sql_ctx, dataset)

    def write(self, df, dataset_name, prefix, **kwargs):
        """When project_key is None use same project as recipe is in."""
        project_key = prefix  # Assume that prefix is a valid project_key
        dataset = dataiku.Dataset(dataset_name, project_key)
        dkuspark.write_with_schema(dataset, df)

    def dataiku_sql_ctx(self):
        sc = pyspark.SparkContext.getOrCreate()
        return SQLContext(sc)



# import pkgutil
import inspect
import os
import sys
import traceback

__all__ = ['exec_code', 'run']


def exec_code(code, globals_dict):
    """Execute the recipe code and handle error conditions.

    Args:
        code (str): The code to be executed.
        globals_dict (dict): globals_dict as used by exec().


    Returns:
       Output of python exec() function.
    """
    e = False
    try:
        ret = exec(code, globals_dict)
    except SyntaxError as err:
        error_class = err.__class__.__name__
        # detail = err.args[0]
        lineno = err.lineno
        e = err
    except Exception as err:
        error_class = err.__class__.__name__
        # detail = err.args[0]
        cl, exc, tb = sys.exc_info()
        lineno = traceback.extract_tb(tb)[-1][1]
        e = err
    if e:
        lines = code.splitlines()
        lenlines = len(lines)
        if lineno < lenlines:
            print("Recipe execution",
                  error_class,
                  "at recipe line:\n",
                  lines[lineno-1])
        else:
            print("Recipe execution",
                  error_class)
        raise e
    return ret


def run(root_mod, recipe, dataframe_source, replacements=[]):
    """Obtain a dataframe. It will adjust to whatever
    storage the environment has set. Currently storage is supported in
    file, memory or dataiku (HDFS).

    Args:
        root_mod (module): The root module on which to base the path.
        recipe (str): Relative path to the recipe file from the module dir.
        dataframe_source (DataframeSourceBase subclass): dataframe source
            E.g. LocalSource, Dataiku
        replacements (list): List of text replacements to enable recipe
        debugging. Example on how to limit data amount:

        [
            {
                "old": "dataframe.get(spark_session, ds_foo.name)",
                "new": "dataframe.get(spark_session, ds_foo.name).limit(10)"
            }
        ]

    Returns:
       Output of python exec() function.
    """
    mod_path = os.path.dirname(inspect.getfile(root_mod))
    recipe_path = f"{mod_path}/{recipe}"
    with open(recipe_path) as f:
        code = prepare_code(f.read(), recipe, replacements)
    globals_dict = {
        'BIRGITTA_DATAFRAMESOURCE': dataframe_source
    }
    return exec_code(code, globals_dict)


def prepare_code(code, recipe, replacements):
    for replacement in replacements:
        code = code.replace(replacement["old"], replacement["new"])
    context_stmts = f"""from birgitta.dataframesource import contextsource
contextsource.set(globals()['BIRGITTA_DATAFRAMESOURCE'])"""
    completed = f"""
print("=== Recipe {recipe} complete ===")"""
    return context_stmts + code + completed



"""Data frame assertion helpers
"""
from pandas.util.testing import assert_frame_equal

__all__ = ['assert_outputs']


def assert_outputs(expected_fixtures, dataframe_source, spark_session):
    """Assert that the expected fixtures are equal to the
    produced outputs. Improves exception presentation."""
    for df_key in expected_fixtures.keys():
        result_df = dataframe_source.load(spark_session, df_key, None)
        expected_df = expected_fixtures[df_key](spark_session)
        result_panda = to_pandas(result_df)
        expected_panda = to_pandas(expected_df)
        # Ugly hack to print schema
        if not result_panda.equals(expected_panda):
            print_schema_diff(df_key, result_df, expected_df)
        try:
            assert_frame_equal(result_panda, expected_panda)
        except AssertionError as e:
            if "Attributes are different" in str(e):
                print("============== Assertion Error:",
                      df_key,
                      "==============")
                print("Result for %s" % (df_key))
                result_df.show()
                print("Expected for %s" % (df_key))
                expected_df.show()
            raise e


def to_pandas(df):
    # Convert timestamp to string to work around arrow error:
    # "Can only use .dt accessor with datetimelike values"
    for field in df.schema.fields:
        ftype = field.dataType.__class__.__name__
        if ftype in ["DateType", "TimestampType"]:
            df = df.withColumn(field.name, df[field.name].cast("string"))
    return df.toPandas()


def print_schema_diff(df_key, result_df, expected_df):
    print_schema(df_key + "_result (left)", result_df)
    print_schema(df_key + "_expected (right)", expected_df)
    print_schema_type_diffs(df_key, result_df, expected_df)


def print_schema_type_diffs(df_key, result_df, expected_df):
    print("============== ", end="")
    print("Data Frame Schema Diff: %s" % (df_key), end="")
    print(" ==============")
    result_fields = schema_to_set(result_df)
    expected_fields = schema_to_set(expected_df)
    not_in_result = sorted(expected_fields.difference(result_fields))
    if not_in_result:
        print("Not in result:", repr(not_in_result))
    not_in_expected = sorted(result_fields.difference(expected_fields))
    if not_in_expected:
        print("Not in expected:", repr(not_in_expected))


def schema_to_set(df):
    field_strs = set()
    for i, f in enumerate(df.schema.fields):
        field_strs.add("%d:%s:%s" % (i, f.name,
                                     f.dataType.__class__.__name__))
    return field_strs


def schema_to_str(df):
    field_strs = []
    for i, f in enumerate(df.schema.fields):
        field_strs.append("%d:%s:%s" % (i, f.name,
                                        f.dataType.__class__.__name__[:8]))
    return ", ".join(field_strs)


def print_schema(name, df):
    print("============== ", end="")
    print("Data Frame Schema: %s" % (name), end="")
    print(" ==============")
    print(schema_to_str(df))



"""Default fixture values for various type of values.
"""
import pandas as pd
from birgitta.schema.fixtures.values import dtvals


def inthepast():
    """We have a hardcoded default timestamp to get consistent
    test execution"""
    return dtvals.format_dt(pd.Timestamp(2019, 1, 1, 1))


def inthepast_str():
    return dtvals.dtstr(inthepast())


def default_timestamp():
    return inthepast()


def default_date():
    return default_timestamp().date()


def today():
    """Today is two months after default timestamp.
    Very useful to get consistent test results."""
    return pd.Timestamp(2019, 3, 1, 1).date()


def today_str():
    return dtvals.today_str(today())


def year_end():
    return pd.Timestamp(2019, 12, 31, 1).date()


def today_datetime():
    """Today is two months after default timestamp"""
    return dtvals.format_dt(pd.Timestamp(2019, 3, 1, 1))


def timestamp(day_delta=0, hour_delta=0, hard_val=None):
    if hard_val:
        if hard_val == "today":
            dt = today()
        else:
            return hard_val
    else:
        dt = default_timestamp() + pd.Timedelta(days=day_delta,
                                                hours=hour_delta)
        dt = dtvals.format_dt(dt)
    return dt


def timestamp_str():
    return dtvals.dtstr(timestamp())


def NaT():
    return None


def nan():
    return None



"""Functions for diffing pyspark dataframes."""

import pandas as pd

__all__ = ['diff', 'assert_equals']


def array_diff(first, second):
    if len(first) != len(second):
        return True
    first = set(first)
    second = set(second)
    return [item for item in first if item not in second]


def cols_diff(expected, actual):
    """Should have same columns"""
    if array_diff(expected.columns, actual.columns):
        expected_cols = ",".join(expected.columns)
        actual_cols = ",".join(actual.columns)
        return """Error: Cols diff
        Expected: %s
        Actual:   %s""" % (expected_cols, actual_cols)
    return None


def dtypes_diff(expected, actual):
    """Should have same data types"""
    if array_diff(expected.dtypes, actual.dtypes):
        print("actual.dtypes:", repr(actual.dtypes))
        expected_dtypes = repr(expected.dtypes)
        actual_dtypes = repr(actual.dtypes)
        return """Error: Column type diff
        Expected: %s
        Actual:   %s""" % (expected_dtypes, actual_dtypes)
    return None


def diff(expected, actual):
    """Diff two dataframes. However, this function has errors
    as it only subtracts expected from actual, when it should
    test the opposite also."""
    col_diff = cols_diff(expected, actual)
    if col_diff:
        return col_diff
    dtype_diff = dtypes_diff(expected, actual)
    if dtype_diff:
        return dtype_diff
    expected_count = expected.count()
    actual_count = actual.count()
    # Same number of rows
    if expected_count != actual_count:
        return """Error: Row count diff
        Expected: %s
        Actual:   %s""" % (expected_count, actual_count)
    # Are the records the same
    expected_pdf = panda_sorted(expected)
    actual_pdf = panda_sorted(actual)
    if not expected_pdf.equals(actual_pdf):
        diff_df = expected.subtract(actual)
        reverse_diff_df = actual.subtract(expected)
        differing_rows_str = panda_sorted(diff_df).to_string()
        reverse_differing_rows_str = panda_sorted(reverse_diff_df).to_string()
        expected_rows_str = expected_pdf.to_string()
        actual_rows_str = actual_pdf.to_string()
        return """Error: Rows are different
        Only in expected:
        %s
        Only in actual result:
        %s
        Expected:
        %s
        Actual:
        %s""" % (differing_rows_str,
                 reverse_differing_rows_str,
                 expected_rows_str,
                 actual_rows_str)
    return False


def panda_assert_df_equals(expected_df, actual_df):
    expected_pdf = panda_sorted(expected_df)
    actual_pdf = panda_sorted(actual_df)
    pd.testing.assert_frame_equal(expected_pdf,
                                  actual_pdf,
                                  check_like=True,
                                  check_dtype=True)


def panda_sorted(df):
    return df.toPandas().sort_values(df.columns).reset_index(drop=True)


def assert_equals(expected, actual):
    # Create richer diff information
    diff_ret = diff(expected, actual)
    if (diff_ret):
        cols = actual.columns
        col_strs = []
        for idx, col in enumerate(cols):
            col_strs.append("%d:%s" % (idx, col))
        print("Columns:", ", ".join(col_strs))
        print(diff_ret)
    # Perform a deeper comparison to catch dtype and other diffs
    panda_assert_df_equals(expected, actual)



"""The context_source module. Helper function to obtain the data frame source
from the context.
"""
from birgitta import context


__all__ = ['get', 'set']


def get():
    return context.get('BIRGITTA_DATAFRAMESOURCE')


def set(dataframe_source):
    return context.set('BIRGITTA_DATAFRAMESOURCE', dataframe_source)



from birgitta import spark as bspark
from birgitta.recipetest import recipetest
from birgitta.schema.spark import to_spark
from dataiku.scenario import Scenario


spark_session = bspark.session()

scenario = Scenario()
src_project_key = 'TRIBUNE'
src_recipe_key = 'compute_bsparkbookexampleresult'
testbench_project_key = 'TRIBUNE_TESTBENCH'
spark_schema = to_spark([{'name': 'letter', 'type': 'string'},
                         {'name': 'number', 'type': 'bigint'}])

test_params = {
    'principal_output_dataset': 'exampledatasetoutput',
    'schemas': {
        'inputs': {
            'exampledataset': spark_schema,
        },
        'outputs': {
            'exampledatasetoutput': spark_schema
        }
    },
    'test_cases': [
        {
            'name': 'correct_count',
            'inputs': {
                'exampledataset': {
                    'rows': [['a', 1], ['b', 2], ['c', 3], ['d', 4]]
                }
            },
            'outputs': {
                'exampledatasetoutput': {
                    'rows': [['a', 1], ['b', 2]]
                }
            }
        }
    ]
}

recipetest.test_recipe(spark_session,
                       scenario,
                       src_project_key,
                       src_recipe_key,
                       testbench_project_key,
                       test_params)



"""Handle global variables used to:

* Set dataframe source
* Collect spark recipe test coverage data
* Specify Today value, to get consistent tests
* Set other global variables
"""
import datetime


class Context():
    TODAY = datetime.date.today()


def reset():
    Context.BIRGITTA_DATAFRAMESOURCE = None
    Context.BIRGITTA_TEST_COVERAGE = {}
    Context.BIRGITTA_SPARK_SESSION_TYPE = "LOCAL"


reset()  # Init context


def get(key, default=None):
    attr = getattr(Context, key)
    return attr if attr else default


def set(key, val):
    setattr(Context, key, val)



"""Various functions for generate date time values.
"""
import calendar
import datetime
import math
from datetime import timezone

import holidays
from birgitta.schema.fixtures import values as v
from dateutil import rrule


no_holidays = holidays.Norway()


def default_val(field):
    """Get the default value for date time named field.

    Args:
        field (str): field name, e.g. 'year4'

    Returns:
        Corresponding value.
    """
    return field_val(field, v.default_date(), v.today())


def field_val(field, dt, today=None):
    if not today:
        today = v.today()

    if field == 'date':
        return datenum(dt)
    elif field == 'period':
        return datenum(dt)
    elif field == 'month':
        return monthnum(dt)
    elif field == 'week':
        return yearweekstr(dt)
    elif field == 'year4':
        return dt.year
    elif field == 'is_leap_year':
        return calendar.isleap(dt.year)
    elif field == 'halfyear_number':
        return roundup_divide_month(dt, 6)
    elif field == 'tertial_number':
        return roundup_divide_month(dt, 4)
    elif field == 'year_quarter':
        return "%s-Q%d" % (dt.strftime("%Y"), quarter(dt))
    elif field == 'quarter_number':
        return quarter(dt)
    elif field == 'year_month_number':
        return dt.strftime("%Y-%m")
    elif field == 'month_number':
        return dt.month
    elif field == 'isoyear4':
        return dt.isocalendar()[0]
    elif field == 'isoweek':
        return dt.isocalendar()[1]
    elif field == 'week_in_year':
        return dt.isocalendar()[1]
    elif field == 'datetimestamp_parsed':
        return date_to_datetime(dt)
    elif field == 'weeks_ago':
        return weeks_between(dt, today)
    elif field == 'months_ago':
        return months_between(dt, today)
    elif field == 'years_ago':
        return years_between(dt, today)
    elif field == 'dates_ago':
        return days_between(dt, today)
    elif field == 'date_iso_short':
        return datenum(dt)
    elif field == 'day_in_week':
        return dt.isocalendar()[2]
    elif field == 'day_in_month':
        return dt.day
    elif field == 'day_in_year':
        return dt.timetuple().tm_yday
    elif field == 'holidayint':
        return is_holiday(dt)
    elif field == 'is_first_day_in_month':
        return dt.day == 1
    elif field == 'is_last_day_in_week':
        return dt.isocalendar()[2] == 7
    elif field == 'is_last_day_in_month':
        return is_last_day_of_month(dt)
    elif field == 'yesterday':
        return yesterday_str(dt)
    elif field == 'today_str':
        return today_str(dt)
    elif field == 'second_yearweek_str':
        return second_yearweek_str(dt)
    elif field == 'yesterday_str':
        return yesterday_str(dt)
    elif field == 'lastmonth':
        return lastmonth(dt)
    return None


def daterange(start_date=None, end_date=None):
    if not start_date:
        start_date = v.default_date()
    if not end_date:
        end_date = v.today()
    for n in range(int((end_date - start_date).days)):
        yield start_date + datetime.timedelta(n)


def daterange_len(start_date=None, end_date=None):
    if not start_date:
        start_date = v.default_date()
    if not end_date:
        end_date = v.today()
    return int((end_date - start_date).days)


def roundup_divide_month(dt, div):
    return int(math.ceil(dt.month/div))


def date_to_datetime(dt):
    return datetime.datetime(dt.year, dt.month, dt.day)


def quarter(dt):
    return roundup_divide_month(dt, 4)


def is_holiday(dt):
    return 1 if dt in no_holidays else 0


def datenum(dt):
    return int(dt.strftime("%Y%m%d"))


def monthnum(dt):
    return int(dt.strftime("%Y%m"))


def yearweekstr(dt):
    return dt.strftime("%dW%U")


def strtodate(dt, date_str):
    return dt.strftime(datetime.datetime.strptime(date_str, '%Y-%m-%d'))


def is_last_day_of_month(dt):
    # calendar.monthrange return a tuple (weekday of first day of
    # the  month, number of days in month)
    last_day_of_month = calendar.monthrange(dt.year, dt.month)[1]
    # here i check if date is last day of month
    return dt == datetime.date(dt.year, dt.month, last_day_of_month)


def days_between(start_date, end_date):
    days = rrule.rrule(rrule.DAILY, dtstart=start_date, until=end_date)
    return days.count()


def weeks_between(start_date, end_date):
    weeks = rrule.rrule(rrule.WEEKLY, dtstart=start_date, until=end_date)
    return weeks.count()


def months_between(start_date, end_date):
    months = rrule.rrule(rrule.MONTHLY, dtstart=start_date, until=end_date)
    return months.count()


def years_between(start_date, end_date):
    years = rrule.rrule(rrule.YEARLY, dtstart=start_date, until=end_date)
    return years.count()


def yesterday(dt):
    return dt - datetime.timedelta(1)


def dtstr(dt):
    return dt.strftime('%Y-%m-%d')


def today_str(dt):
    return dtstr(dt)


def second_yearweek(dt):
    return dt + datetime.timedelta(8)


def second_yearweek_str(dt):
    return dtstr(second_yearweek(dt))


def yesterday_str(dt):
    return dtstr(yesterday(dt))


def lastmonth(dt):
    return dt - datetime.timedelta(30)


def format_dt(dt):
    if type(dt) is datetime.datetime:
        pydt = dt
    else:
        pydt = dt.to_pydatetime()
    return set_timezone(pydt)


def set_timezone(pydt):
    return pydt.replace(tzinfo=timezone.utc)


def date_types_to_str(val):
    """Convert data types to string. Used to avoid conversion bugs."""
    if type(val) == datetime.date:
        return val.strftime('%Y-%m-%d %H:%M:%S')
    if type(val) == datetime.datetime:
        return val.strftime('%Y-%m-%d')
    return val



"""Module for test case fixturing functions.
"""
from birgitta import timing


__all__ = ['obtain_fixture_fns', 'write_fixtures']


def obtain_fixture_fns(datasets, fixture_name):
    """Obtain fixture functions

    Returns:
        Dict of fixture name -> fixture fn.
    """
    fixture_fns = {}
    for ds in datasets:
        ds_module = ds[0]
        fixture_module = ds[1]
        fixture_fn = getattr(fixture_module, "fx_" + fixture_name)
        fixture_fns[ds_module.dataset.name] = fixture_fn
    return fixture_fns


def write_fixtures(globals_dict, fixture_fns, dataframe_source, spark_session):
    """Write fixtures to storage.

    Args:
        globals_dict (dict): The set of variables which will be
        available to the pyspark recipe as globals. It is needed
        in order to store the memory based fixtures in globals.
        fixture_fns
    """
    timing.time("write_fixtures start")
    # FUTURE: Try writing directly to parquet using pyarrow instead. Could be faster. # noqa 501
    for dataframe_key in fixture_fns.keys():
        df = fixture_fns[dataframe_key](spark_session)
        dataframe_source.write(df, dataframe_key)
    timing.time("write_fixtures end")



"""Field / feature catalog class. This catalog is per project/domain.
It holds a list of fields with description and example values.
It does not hold field types. The field types themselves are defined
in specific data set schemas, for clarity and more intuitive schema APIs.
Additionally, by decoupling data type from field description and example_val,
The same example value e.gg. 2017-01-01 can be interpreted as date, datetime
or string, depending on the schema needs.

The purpose of the catalog is:

1. generate example fixtures for recipe tests in an efficient manner
2. reuse default values across data set fixtures for simplicity
3. promote reuse of field names across data sets
4. document all fields and eventually enable automatic publishing of these
comments in views and reports
"""
from birgitta.schema.fixtures.values import dtvals

__all__ = ['Catalog']


class Catalog:
    def __init__(self, field_confs, secondary_field_confs=None):
        """
        Args:
            field_confs (dict): holds all fields with
            descriptions and example values.
            secondary_field_confs (dict): fall back dict enabling
            the combination of fields from separate domains.
        """
        self.field_confs = field_confs
        self.secondary_field_confs = secondary_field_confs

    def example_val(self,
                    field,
                    override_conf=None):
        """Get the example val for a field.

        Args:
            field (str): Field name.
            override_conf (dict): Overrides catalog fixture conf.

        Returns:
            example value
        """
        example_conf = self.get_example_conf(field, override_conf)
        return self.val_from_conf(field, example_conf)

    def get_example_conf(self, field, override):
        """Get theh example conf for a field.

        These are the possible sources of example confs:
        Default is the conf from the catalog.
        If there is an override, it is used instead.
        In this way, a more specific value can be used for specific cases.
        If none is found, then fall back to default date time val.
        This is useful for building date dimension tables.
        """
        conf = None
        if override:  # overrides all
            conf = override
        else:  # Look for a conf in the catalog
            conf = self.catalog_conf(field)
        if conf:
            return conf["example"]
        # Interpret field as generic datetime value if no conf found,
        # enabling writing schemas faster for date dimension tables.
        return {"dtval": field}

    def val_from_conf(self, field, example_conf):
        # If dtval, get standard time based value from dtvals lib.
        if "dtval" in example_conf:  # Use dtvals
            return dtval(example_conf)
        # If fn val, invoke the value generating function
        if "fn" in example_conf:
            return fnval(example_conf)
        # If static val, return it
        if "static" in example_conf:
            return staticval(field, example_conf)

    def catalog_conf(self, field):
        if field in self.field_confs:
            return self.field_confs[field]
        # If not in primary field_confs, test secondary_field_confs
        elif (self.secondary_field_confs and
              (field in self.secondary_field_confs)):
            return self.secondary_field_confs[field]
        return None


def dtval(example_conf):
    """Get a standard timebased value from the dtvals lib."""
    dt_field = example_conf['dtval']
    if "params" in example_conf:
        return dtvals.field_val(dt_field, example_conf["params"]["dt"])
    else:
        return dtvals.default_val(dt_field)


def fnval(example_conf):
    fn = example_conf["fn"]
    if "params" in example_conf:
        return fn(*list_if_not(example_conf["params"]))
    else:
        return fn()


def staticval(field, example_conf):
    val = example_conf["static"]
    if callable(val):
        raise ValueError(
            '"static" field %s cannot be callable, use "fn": %s' % (
                field, repr(val)))
    return example_conf["static"]


def list_if_not(params):
    if isinstance(params, list):
        return params
    else:
        return [params]



"""The main package for the Python Module Distribution Utilities.
A normal use is to load a dataframe in pyspark code. By loading
the dataframe this way, fixture testing of the notebook is enabled:

   from birgitta.dataframe import dataframe
   dataframe.get(spark_session, 'dataset_name')
"""



"""The dataframesource module. Provides classes for reading and writing
dataframes from different sources.
"""

from . import contextsource
from .dataframesourcebase import DataframeSourceBase


__all__ = ['DataframeSourceBase', 'sources', 'contextsource']  # noqa F822



"""The dataframesource abstract base class.
"""

from abc import ABCMeta, abstractmethod

__all__ = ['DataframeSourceBase']


class DataframeSourceBase(metaclass=ABCMeta):
    def __init__(self, **kwargs):
        """Abstract data fram source class.

        Args:
            **kwargs (keyword args): optional keyword args for subclasses.
            e.g. dataset_dir for LocalSource.
        """
        self.keyword_args = kwargs

    @abstractmethod
    def load(self, spark_session, dataset_name, prefix, **kwargs):
        """Method for getting a data frame.

        Args:
            dataset_name (str): Name of the data set.
            prefix (str): Prefix path or dataiku project_key for loading
            the data set.
            sqlContext (SQLContext): SQLContext needed to load or
            instantiate the data frame.

        Returns:
            A data frame of the data set."""
        return NotImplemented

    @abstractmethod
    def write(self, df, dataset_name, prefix, **kwargs):
        """Write a data frame to a data source.

        Args:
            df (DataFrame): The DataFrame to write.
            dataset_name (str): Name of the data set.
            prefix (str): Prefix path or dataiku project_key for loading
            the data set.
."""
        return NotImplemented



"""The Schema class. Used to handle fixtures and managed spark
schemas.
"""
from birgitta.schema.spark import enforce_schema, to_spark, to_escaped_spark  # noqa 401

NAME_POS = 0
TYPE_POS = 1

__all__ = ['Schema']


class Schema:
    def __init__(self, arr_schema, catalog):
        """Constructs a schema object which should be treated
        as immutable.

        Args:
            arr_schema (list): a schema in the list based simple schema DSL.
            catalog (Catalog): the field catalog to get fixtures and
            descriptions from.
        """
        self.arr_schema = arr_schema
        self.catalog = catalog

    def to_spark(self):
        return to_spark(self.arr_schema)

    def to_escaped_spark(self):
        return to_escaped_spark(self.arr_schema)

    def fields(self):
        return list(map(lambda f: f[NAME_POS], self.arr_schema))

    def types(self):
        return self.dict()

    def enforce(self, df):
        return enforce_schema(df, self)

    def example_val_override(self, field):
        return self.field_params()[field]

    def field_params(self):
        ret = {}
        for row in self.arr_schema:
            field = row[0]
            if len(row) > 2:
                ret[field] = row[2]
            else:
                ret[field] = {}
        return ret

    def dict(self):
        return {k[NAME_POS]: k[TYPE_POS] for k in self.arr_schema}



from os import path

from setuptools import find_packages
from setuptools import setup


version = '0.1.9'
here = path.abspath(path.dirname(__file__))

long_description = """Birgitta is a Python ETL test and schema framework,
providing automated tests for pyspark notebooks/recipes.

Birgitta allows doing solid ETL and ML, while still liberally
allowing imperfect notebook code, enabling a
`DataOps <https://www.dataopsmanifesto.org>`_ way of
working which is both solid and agile, not killing
Data Scientist flexibility by excessive coding standards in notebooks.

In addition to running recipetests on your local dev machine or on
a CI/CD server, there is support for running recipetests
as [Dataiku](https://www.dataiku.com] DSS Scenarios.
"""

setup(
    name='birgitta',

    version=version,

    description='Pyspark and notebook unit testing, especially focused on Dataiku.', # noqa 501
    long_description=long_description,

    url='https://github.com/telia-oss/birgitta',

    author='Telia Norge',

    license='MIT',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ],

    keywords='',

    packages=find_packages(exclude=[
        'contrib',
        'docs',
        '*.tests',
        '*.tests.*',
        'tests.*',
        'tests'
        ]),

    # List of dependencies, with exact versions
    install_requires=[
        'docopt',
        'numpy',
        'pyspark',
        'mock',
        'pandas',
        'pyarrow',
        'pytest',
        'pytest-cov',
        'pytest_mock',
        'holidays'
    ],

    include_package_data=True,
    package_data={},
    entry_points={
        'console_scripts': [
        ]
    }
)



"""The S3 dataframesource class.
"""
from birgitta.dataframesource import DataframeSourceBase

__all__ = ['S3Source']


class S3Source(DataframeSourceBase):
    """Instantiate an S3Source object, defining the format, e.g. 'json'
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if 'format' in kwargs:
            self.format = kwargs['format']
        else:
            self.format = 'parquet'

    def load(self, spark_session, dataset_name, prefix, **kwargs):
        s3_path = prefix  # Assume that prefix is a valid s3_path
        dest_path = f"s3a://{s3_path}/{dataset_name}.{self.format}"
        return spark_session.read.format(self.format).load(dest_path)

    def write(self, df, dataset_name, prefix, **kwargs):
        s3_path = prefix  # Assume that prefix is a valid s3_path
        dest_path = f"s3a://{s3_path}/{dataset_name}.{self.format}"
        df.write.format(self.format).mode("overwrite").save(dest_path)
